import argparse

# parser = argparse.ArgumentParser(description="A command line interface for the banditvis module.")

def get_args():
    parser = argparse.ArgumentParser(prog='banditvis')
    parser.add_argument("input",
        help="The path to your .txt input file.")
    parser.add_argument("-d", "--data", nargs='?',
        help="The path to the data directory, defaults to the current directory.")
    parser.add_argument("-o", "--out", nargs='?',
        help="The path to the output location, defaults to the current directory.")
    args = parser.parse_args()
    print(vars(args))
    return vars(args)


import numpy as np


class BernoulliArm:
    """
    Arm with reward probabilty from a Bernoulli distribution.

    A BernoulliArm is called with a single argument: the mean (p). It returns a
    reward of 1 with probability p, and a reward of 0 with probability (1 - p).
    """
    def __init__(self, attr_list):
        self.mean = attr_list[0]
        self.info = ["Bernoulli", self.mean]

    def pull(self):
        if np.random.random() > self.mean:
            return 0
        return 1


class NormalArm:
    """
    A NormalArm selects rewards from a normal distribution.

    The mean and variance are provided during initialization.
    """
    def __init__(self, attr_list):
        self.mean = attr_list[0]
        self.variance = attr_list[1]
        self.info = ["Normal", self.mean, self.variance]

    def pull(self):
        return np.random.normal(self.mean, np.sqrt(self.variance))


class LinearArm:
    """
    A LinearArm selects rewards as a linear function of the arm and the mean.

    The arm vector and mean vector are provided as arguments; the reward from
    pulling a certain arm is given by the inner product of that vector and the
    mean vector plus some scalar sub-gaussian noise.

    Note that in the LinBandit declaration, the mean vector is passed to the
    LinearArm for use as the mean_vec.
    """
    def __init__(self, arm_vector):
        self.arm_vec = arm_vector
        self.dim = len(arm_vector)
        self.info = ["Linear", self.arm_vec]

    def pull(self):
        return np.inner(self.arm_vec, self.mean_vec) + np.random.normal(0,1)


from sys import stdout
import time
from .algorithms import *
from .arms import *
from .core import *

from pprint import pprint
class Simulation:

    """
    The Simulation class combines the Bandit - Algorithm interaction.

    It stores various information about the interaction, and it contains
    various execution methods which are useful depending on the context, and
    various file output methods that are called by other files that need to
    simulate a bandit and access the saved data.

    TODO:
    * what else to track? total reward, calculate regret, ...

    """

    def __init__(self, bandit_inst, algorithm_inst):
        """
        Initializes the Simulation class, and passes certain attributes so that
        Bandit objects and Algorithm objects have acces to certain information.
        """
        self.alg = algorithm_inst
        self.bandit = bandit_inst
        self.alg.bandit = bandit_inst
        self.total_regret = 0
        self.iterations = 0
        self.arm_info = [each_arm.info for each_arm in self.bandit.arms]


    def reset(self):
        """
        Resets the Simulation class the Bandit subclass.
        """
        self.iterations = 0
        self.total_regret = 0
        self.bandit.reset()


    def _print_info(self):
        """
        Prints out information about the Simulation.

        The 'try' statement attempts to print out the .__name__ attribute of
        functions, rather than the object itself; if it fails, it falls back to
        just printing the object itself
        """
        print("\nAverage Regret: {0}".format(
            self.total_regret / self.iterations))

        for i in self.alg.var_dict:
            try:
                print("{0} : {1}".format(i, self.alg.var_dict[i].__name__))
            except AttributeError:
                print("{0} : {1}".format(i, self.alg.var_dict[i]))

        print("{0} Arms: {1}".format(self.bandit.n_arms, self.arm_info))
        print("Horizon: {0}".format(self.horizon))
        print ("Total runtime: {0} seconds".format(self.runtime))
        print ("Average runtime per iteration: {0} seconds".format(
            self.runtime / self.iterations))
        print("\n" + "-"*50 + "\n")
        return None


    def runStep(self, num, horizon):
        """
        Runs the Simulation for one step.

        This method is called during animations.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for m in range(num):
            self.bandit.pullArm(self.alg.giveArm())
        return None


    def runView(self, horizon):
        """
        Runs a single iteration to a horizon of n and prints out information.

        runView is especially useful for debugging bandit algorithms or seeing
        everything in significant detail as the bandit runs.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        start_time = time.clock()
        for m in range(self.bandit.n_arms):
            self.bandit.pullArm(m)
            self.bandit.fullInfo()
        for m in range(horizon - self.bandit.n_arms):
            self.bandit.pullArm(self.alg.giveArm())
            self.bandit.fullInfo()
        stop_time = time.clock()

        self.total_regret += self.bandit.giveRegret()
        self.iterations += 1
        self.bandit.reset()
        self.runtime = stop_time - start_time
        self._print_info()
        return None


    def runStandard(self, horizon, cycles):
        """
        A standard experiment-running method.

        Runs the bandit for a certain number of cycles each to a certain
        horizon, and prints out the result on screen.

        It uses _print_info to print information about the bandit itself upon
        termination.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        print("-"*50 + "\n")
        start_time = time.clock()

        for i in range(cycles):
            for i in range(self.bandit.n_arms):
                self.bandit.pullArm(i)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1
            self.bandit.reset()

            stdout.write(
                "\r----------  "
                "{0:d} out of {1} cycles"
                "----------".format(self.iterations, cycles))
            stdout.flush()

        stop_time = time.clock()
        self.runtime = stop_time - start_time
        self._print_info()
        return None


    def runHist(self, horizon, cycles, out_file):
        """
        Produces Histogram - like data.

        Every cycle, prints the bandit regret it out in the file, followed by a
        newline. These data files are used by many other methods containing
        hist in their name.
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for i in range(cycles):
            for j in range(self.bandit.n_arms):
                self.bandit.pullArm(j)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1

            with open(out_file, "a") as outfile:
                outfile.write("{0}\n".format(self.bandit.giveRegret()))
            self.bandit.reset()
        return None


    def runVar(self, horizon, cycles, out_file):
        """
        Produces Variable - like data.

        Runs the simulation for a certain number of cycles, then appends that
        data to the text file
        """
        self.bandit.horizon = np.full(self.bandit.n_arms, horizon, dtype = int)
        self.horizon = horizon

        for i in range(cycles):
            for i in range(self.bandit.n_arms):
                self.bandit.pullArm(i)

            for l in range(horizon - self.bandit.n_arms):
                self.bandit.pullArm(self.alg.giveArm())

            self.total_regret += self.bandit.giveRegret()
            self.iterations += 1

            self.bandit.reset()

        with open(out_file, "a") as outfile:
            outfile.write("{0}\n".format(self.total_regret/self.iterations))
        return None

    def selfCheck(self):
        """
        Checks the simulation for errors
        """
        errors = ""
        return errors

class ObjectDict:
    ArmDict = {
        'Bernoulli' : BernoulliArm,
        'Normal' : NormalArm,
        'Linear' : LinearArm
    }
    AlgDict = {
        'var_dict' : {},
        'random' : random,
        'greedy' : greedy,
        'greedy_ep' : greedy_ep,
        'UCB' : UCB,
        'UCB_KL' : UCB_KL,
        'TS_Beta' : TS_Beta,
        'TS_Gauss' : TS_Gauss,
        'Bayes_Gauss' : Bayes_Gauss,
        'UCB_Lin' : UCB_Lin,
        'TS_Lin' : TS_Lin
    }
    IndexDict = {
        'B1' : B1,
        'B2' : B2,
        'B3' : B3,
        'B4' : B4,
        'B5' : B5,
        'B6' : B6,
        'B7' : B7,
        'B8' : B8,
        'B9' : B9
    }


def ReMapSim(sim_dict):
    """
    Takes a simulation dictionary and turns string names into objects. It uses
    the ObjectDict class in order to create references to the actual objects.
    """
    for alg_key in list(sim_dict['Algorithm'].keys()):
        if alg_key == 'incr':
            sim_dict['Algorithm'][alg_key] = \
                ObjectDict.IndexDict[sim_dict['Algorithm'][alg_key]]
        elif alg_key == 'algtype':
            sim_dict['Algorithm'][alg_key] = \
                ObjectDict.AlgDict[sim_dict['Algorithm'][alg_key]]

    sim_dict['Algorithm'] = Algorithm(**sim_dict['Algorithm'])
    sim_dict['arm_object_list'] = [ObjectDict.ArmDict[arm[0]](arm[1])
        for arm in sim_dict['Bandit']['ArmList']]

    # different parsing for Linear and non-Linear bandits
    if sim_dict['Bandit']['ArmList'][0][0] == 'Linear':
        sim_dict['vector_mean'] = sim_dict['Bandit']['MeanVector']
        sim_dict['Bandit'] = LinBandit(
            sim_dict['arm_object_list'],
            sim_dict['vector_mean'],
            normalized=sim_dict['Normalized'])
        del[sim_dict['vector_mean']]
    else:
        sim_dict['Bandit'] = Bandit(sim_dict['arm_object_list'])

    sim_dict['Simulation'] = Simulation(
        sim_dict['Bandit'],
        sim_dict['Algorithm'])

    # clean up residual keys
    del[sim_dict['Bandit']]
    del[sim_dict['Algorithm']]
    del[sim_dict['arm_object_list']]

    return None



import os

import matplotlib.pyplot as plt
import numpy as np

from .data import VarData
from .simulation import ReMapSim
from .formatting import cmap_colors, mpl_defaults


def _safe_plot_save(file_name):
    """Save a plot safely."""
    if file_name.split("/")[-1][:4] == 'temp':
        plt.savefig(file_name)
    elif not os.path.isfile(file_name):
        plt.savefig(file_name)
    else:
        for i in range(100000):
            if os.path.isfile(file_name.split(".")[0]
                + "{}.".format(i)
                + file_name.split(".")[1]):
                pass
            else:
                plt.savefig(file_name.split(".")[0]
                    + "{}.".format(i)
                    + file_name.split(".")[1])
                break
    return None


def VarPlot(core_dict):
    """
    Makes a Variable plot out of existing data.

    The data is loaded from the data folder, whose location is specified in the
    input dictionary. The loop iterates over files in that folder, and for each
    one adds a plot to the figure.

    Uses _safe_plot_save to save the plot when finished.

    TODO:
    * add proper label support from user input
    """

    # defaults
    fig, ax = mpl_defaults.plot()

    # add plots
    for i, sub_dict in enumerate(core_dict['sim']):
        y_list = np.loadtxt(
            "{}/data{}.txt".format(core_dict['DataFolder'], i),
            float)
        x_list = core_dict['arg_list']

        cmap = cmap_colors.sequential1[i]

        plt.plot(x_list, y_list,".--",
            linewidth=1,
            color = cmap(0.8),
            label = sub_dict['label'])

    # labels / text
    legend = plt.legend(loc='best', framealpha = 1.0)
    legend.get_frame().set_linewidth(1)
    plt.xlabel(core_dict['xlabel'])
    plt.ylabel(core_dict['ylabel'])
    plt.title(core_dict['PlotTitle'], style='italic')

    # save plot
    _safe_plot_save(core_dict['PlotSave'])

    return None


def HistPlot(core_dict):
    """
    Makes a Histogram plot out of existing data.

    The data is loaded from the data folder, whose location is specified in the
    input dictionary. The loop iterates over files in that folder, and for each
    one adds a plot to the figure.

    Uses _safe_plot_save to save the plot when finished.

    TODO:
    * move HistData out of this function
    * add support for custom user (x,y)labels, if so desired
    """

    # defaults
    fig, ax = mpl_defaults.plot()

    # variables
    bins = np.linspace(0, np.amax(core_dict['bins']), num = 90)

    # add plots
    for i, sub_dict in enumerate(core_dict['sim']):
        data = np.loadtxt(
            "{}/data{}.txt".format(core_dict['DataFolder'], i),
            float)

        cmap = cmap_colors.sequential1[i]

        avgline = np.average(data)
        plt.axvline(x = avgline, color = cmap(0.5),ls="--" ,linewidth = 1.7)

        plt.hist(data, bins,
            normed=1,
            alpha=0.6,
            facecolor = cmap(0.8),
            label = sub_dict['label'])

    # labels / text
    legend = plt.legend(loc='best', framealpha = 1.0)
    legend.get_frame().set_linewidth(1)
    plt.xlabel('Regret')
    plt.ylabel('Frequency')
    plt.title(core_dict['PlotTitle'], style='italic')

    # save plot
    _safe_plot_save(core_dict['PlotSave'])

    return None


import copy

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib.patches as mpatches
from scipy.stats import beta

from .simulation import ReMapSim
from .formatting import cmap_colors, mpl_defaults


def HistAnimation(core_dict):
    """
    Makes an animation of a Histogram.

    TODO
    * fix bins
    * custom user labels
    * add support for custom fps
    """

    # defaults
    fig, ax = mpl_defaults.ani()

    plt.rcParams['figure.titlesize'] = 14
    plt.rcParams['font.size'] = 12
    plt.rcParams['axes.labelsize'] = 13

    # plot variables
    cmap1 = plt.cm.get_cmap('BuGn')
    cmap2 = plt.cm.get_cmap('Blues')
    bins = np.linspace(0,core_dict['bins'][0],100)

    # formatting
    plt.tick_params(
        axis='y',
        which='both',
        left='off',
        right='off')
    plt.tick_params(
        axis='x',
        which='both',
        top='off')
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.spines['left'].set_color('none')

    # load data
    data = np.loadtxt(core_dict['DataFolder'] + "/data0.txt", float)

    # update function used by the animation
    def _update(num, data):
        nonlocal cmap1, bins, ax

        # clear axes, load data to refresh
        plt.cla()
        data = np.loadtxt(core_dict['DataFolder'] + "/data0.txt", float)

        # plots
        plt.axvline(x = np.average(data),
            color = cmap1(0.5),
            ls="--" ,
            linewidth=1.7)
        plt.hist(data, bins,
            alpha=0.6,
            normed=1,
            facecolor=cmap1(0.8),
            label="X ~ Beta(2,5)")

        # labels
        legend = plt.legend(loc='upper right', framealpha = 1.0)
        legend.get_frame().set_linewidth(1)
        plt.title(core_dict['PlotTitle'], style='italic')

        plt.xlabel('Regret')
        plt.ylabel('Frequency')
        ax.set_ylim([0,0.2])
    # build the animation and run it
    my_ani = animation.FuncAnimation(fig, _update,
        fargs=(data, ),
        interval=1000/core_dict['FPS'])
    plt.show()


def ConfAnimation(core_dict):
    """
    Animates the confidence intervals.

    Since only one cycle is run at a given time, the bandit algorithm is run at
    the same time as the animation rather than using an external data source.

    TODO
    * y limits might not be correct in some cases; specify
    """

    # defaults
    fig, ax = mpl_defaults.ani()

    # map the core_dict to actual objects
    ReMapSim(core_dict['sim'][0])

    # plot variables
    sim = core_dict['sim'][0]['Simulation']
    arms = sim.bandit.n_arms
    title = "Upper Confidence Animation\n"
    not_picked_color = plt.cm.get_cmap('BuGn')(0.6)
    picked_color = plt.cm.get_cmap('OrRd')(0.6)
    bar_width = 0.45

    # simulation variables
    horizon = core_dict['horizon']
    confidence = sim.bandit.U_conf
    index = np.arange(arms)

    # initialize the bandit
    for i in range(arms):
        sim.bandit.pullArm(i)

    # the animation update function
    def _update(num, confidence):
        nonlocal ax

        # ste bandit 1 timestep
        sim.runStep(1, horizon)
        confidence = sim.bandit.U_conf
        max_confidence = np.amax(sim.bandit.U_conf)

        # clear plot to redraw
        plt.cla()
        x_labels = ["Pulls: "+str(item) for item in sim.bandit.T]
        # make a bar plot
        bar_list = plt.bar(index, confidence, bar_width,
            align='center',
            color=not_picked_color,
            edgecolor='#e6e6e6')

        # add a maximum line
        plt.axhline(y=max_confidence,
            ls='dashdot',
            xmin=0.0,
            xmax=1.0,
            linewidth=2.0,
            color='#434343')

        # change the colour of the best arm (the picked arm)
        bar_list[np.argmax(confidence)].set_facecolor(picked_color)

        # add dashed lines for the actual mean of each arm
        for i in range(sim.bandit.n_arms):
            x = [i - bar_width/2, i + bar_width/2]
            y = [sim.bandit.mean_list[i], sim.bandit.mean_list[i]]
            plt.plot(x,y,'--',
                linewidth = 2,
                color='#000000',
                ls='dashed',
                mfc='white')

        # add dotted lines for the empirical mean of each arm
        for i in range(sim.bandit.n_arms):
            x = [i - bar_width/2, i + bar_width/2]
            y = [sim.bandit.U[i], sim.bandit.U[i]]
            plt.plot(x,y,'--',
                linewidth = 1,
                color='#000000',
                ls='-',
                mfc='white')

        # add x labels
        plt.xticks(index, x_labels, weight='bold')

        # add two legends:
        #   * a real legend
        #   * one to display interesting information
        picked_color_path = mpatches.Patch(
            color=picked_color,
            label='Arm chosen')
        not_picked_color_path = mpatches.Patch(
            color=not_picked_color,
            label='Arm(s) not chosen')
        dash_label, = plt.plot([], [],
            ls="dashed",
            label="Actual Mean",
            color='black')
        dashdot_label, = plt.plot([], [],
            ls="-",
            linewidth=1.0,
            label="Empirical Mean",
            color='black')
        legend_1 = plt.legend(
            handles=[picked_color_path, not_picked_color_path, dash_label, dashdot_label],
            loc='upper right')
        plt.gca().add_artist(legend_1)  # prevent legend_2 from overwriting
        legend_1.get_frame().set_linewidth(1)

        blank_path_1 = mpatches.Patch(
            label='Regret: {:04.2f}'.format(sim.bandit.giveRegret()))
        blank_path_2 = mpatches.Patch(
            label='Timestep: {:2d}'.format(sim.bandit.timestep[0]))
        legend_2 = plt.legend(
            handles=[blank_path_1, blank_path_2],
            loc='upper left',
            handlelength=0,
            handletextpad=0)
        legend_2.get_frame().set_linewidth(1)

        # formatting
        ax.grid(False)  # resets grid default
        ax.set_ylim([0,1.2])
        ax.yaxis.grid(True)  # turns horizontal grid on

        plt.title(title)

    # the animation declaration
    my_ani = animation.FuncAnimation(fig, _update,
        fargs=(sim.bandit.U_conf, ),
        interval=1000/core_dict['FPS'])  # fargs is used as the data required in update_hist

    plt.show()


def EllipseAnimation(core_dict):
    """
    Animates the confidence intervals.

    Since only one cycle is run at a given time, the bandit algorithm is run at
    the same time as the animation rather than using an external data source.

    TODO
    * y limits might not be correct in some cases; specify
    """

    # defaults
    fig, ax = mpl_defaults.ani()
    ax.set_aspect('equal')
    # map the core_dict sim to actual objects
    ReMapSim(core_dict['sim'][0])

    # simulation variables
    sim = core_dict['sim'][0]['Simulation']
    sim_dict = core_dict['sim'][0]
    arms = sim.bandit.arm_vecs
    mean = np.array(sim.bandit.mean)
    horizon = core_dict['horizon']
    title = "2D Confidence Ellipse Animation\n"

    # the locations of the perpindicular projection of the arm vector onto the
    # mean vector
    projs = np.array(
        [mean / np.linalg.norm(mean)**2 * np.inner(arm, mean) for arm in arms])
    cmap1 = plt.cm.get_cmap('BuGn')
    cmap2 = plt.cm.get_cmap('Reds')

    # establish plot boundaries
    all_points = np.vstack((arms, mean))
    x_max = np.amax(all_points[:,0])
    x_min = np.amin(all_points[:,0])
    y_max = np.amax(all_points[:,1])
    y_min = np.amin(all_points[:,1])
    x_margin = (x_max-x_min)
    y_margin = (y_max-y_min)
    x_max += x_margin/3
    x_min -= x_margin/5  # less margin on left
    y_max += y_margin/3
    y_min -= y_margin/5  # less margin at bottom

    # set up the figure and axes
    # fig = plt.figure(figsize=(14,9), facecolor='white')
    # ax = fig.add_subplot(111, axisbg='white')

    # initialize the bandit
    for i in range(sim.bandit.n_arms):
        sim.bandit.pullArm(i)


    def plot_ellipse(rho, cov, pos, ax=None, **kwargs):
        """
        Plots an ellipse based on the specified covariance on the parameter rho
        matrix (`cov`). Additional keyword arguments are passed on to the
        ellipse patch artist.

        Parameters
        ----------
            rho : the radius of the ellipse
            cov : The 2x2 covariance matrix to base the ellipse on
            pos : The location of the center of the ellipse. Expects a 2-element
                sequence of [x0, y0].
            ax : The axis that the ellipse will be plotted on. Defaults to the
                current axis.
            Additional keyword arguments are pass on to the ellipse patch.

        Returns
        -------
            A matplotlib ellipse artist
        """
        def eigsorted(cov):
            vals, vecs = np.linalg.eigh(cov)
            order = vals.argsort()[::-1]
            return vals[order], vecs[:,order]

        if ax is None:
            ax = plt.gca()

        vals, vecs = eigsorted(cov)
        theta = np.degrees(np.arctan2(*vecs[:,0][::-1]))

        # width and height are full width, not radius
        width, height = 2 * np.sqrt(rho * vals)
        ellip = mpatches.Ellipse(
            xy=pos,
            width=width,
            height=height,
            angle=theta,
            **kwargs)
        ax.add_artist(ellip)
        return ellip

    # the animation update function
    def _update(num, confidence):
        nonlocal ax

        # step the bandit
        sim.runStep(1, horizon)
        rho = sim.bandit.dim * np.log(sim.bandit.timestep[0])
        G_inv = np.linalg.inv(sim.bandit.G)
        bandit_mean = sim.bandit.U
        chosen_arm = sim.bandit.arm_vecs[np.argmax(sim.bandit.U_conf)]

        # clear plot to redraw
        plt.cla()

        # general formatting
        ax.grid(True)
        ax.set_ylim([y_min,y_max])
        ax.set_xlim([x_min,x_max])
        ax.set_aspect('equal')
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_bbox(dict(
                facecolor='white',
                edgecolor='None',
                alpha=0.65 ))

        # format the axes
        ax.spines['left'].set_visible(True)
        ax.xaxis.set_ticks_position('bottom')
        ax.spines['bottom'].set_position(('data',0))
        ax.yaxis.set_ticks_position('left')
        ax.spines['left'].set_position(('data',0))

        # remove label option
        if core_dict['NoAxesTick']:
            ax.set_xticks([])
            ax.set_yticks([])

        # adds circle for normalized bandit
        if sim_dict['Normalized']:
            ax.add_artist(mpatches.Circle(
                xy=(0.,0.),
                radius=1,
                edgecolor='black',
                alpha=0.2,
                facecolor='none'))

        # position of mean vector
        plt.plot([mean[0]],[mean[1]], "o",
            color='black',
            label='True Mean')

        # plot the ellipse and the mean; checks for LevelCurves option
        if (sim.alg.var_dict['algtype'].__name__ == "TS_Lin"
            and core_dict['LevelCurves'] == True):
            for i in [33, 29, 25, 21, 17, 13, 9, 5, 1]:
                plot_ellipse(i, G_inv, bandit_mean,
                    alpha=0.1,
                    ls='dashed',
                    facecolor=cmap1(0.2),
                    edgecolor='black')
        else:
            plot_ellipse(rho, G_inv, bandit_mean,
                alpha=0.2,
                facecolor=cmap1(0.5),
                edgecolor=cmap1(1.0))
        plt.plot([sim.bandit.U[0]], [sim.bandit.U[1]], "o",
            markerfacecolor=cmap2(0.5),
            markeredgecolor=cmap2(1.0),
            label='Approximated Mean')

        # HelpLines: dashed projection lines to show reward value
        if core_dict['HelpLines']:
            for proj, arm in zip(projs, arms):
                plt.plot([proj[0], arm[0]], [proj[1], arm[1]],
                    color='black',
                    linestyle='dashdot',
                    linewidth=0.8)
            plt.plot([-mean[0]*50,mean[0]*50],[-mean[1]*50,mean[1]*50], "--",
                linewidth=0.5,
                markersize=3,
                color='black')

        # arm vector points
        plt.plot([arm[0] for arm in sim.bandit.arm_vecs],
            [arm[1] for arm in sim.bandit.arm_vecs], "o",
            markersize=5,
            markerfacecolor=cmap1(0.5),
            markeredgecolor=cmap1(1.0),
            label='Arm Vectors')

        # chosen arm; draws a red circle around it
        plt.plot(chosen_arm[0],chosen_arm[1], "or",
            markersize=10,
            markerfacecolor='none',
            markeredgecolor='red',
            markeredgewidth=2,
            label='Chosen Arm')

        # legend and title
        legend = plt.legend(loc='upper right')
        plt.gca().add_artist(legend)
        legend.get_frame().set_linewidth(1)

        # additional information legend
        blank_path_1 = mpatches.Patch(
            label='Regret: {:04.2f}'.format(sim.bandit.giveRegret()))
        blank_path_2 = mpatches.Patch(
            label='Timestep: {:2d}'.format(sim.bandit.timestep[0]))
        legend_2 = plt.legend(
            handles=[blank_path_1, blank_path_2],
            loc='upper left',
            handlelength=0,
            handletextpad=0)
        legend_2.get_frame().set_linewidth(1)

        plt.title(title)

    # the animation declaration
    my_ani = animation.FuncAnimation(fig, _update,
        fargs=(sim.bandit.U_conf, ),
        interval=1000/core_dict['FPS'])  # fargs is used as the data required in update_hist

    plt.show()


def DistAnimation(core_dict):
    """
    Creates an animation of the distributions used for beta and gaussian
    random sampling and confidence algorithms.
    """
    # defaults
    fig, ax = mpl_defaults.ani()


    ReMapSim(core_dict['sim'][0])

    # simulation variables
    sim = core_dict['sim'][0]['Simulation']
    sim_dict = core_dict['sim'][0]
    horizon = core_dict['horizon']
    n_arms = sim.bandit.n_arms
    title = "Distribution Animation\n"
    if sim.alg.var_dict['algtype'].__name__[-5:] == 'Gauss':
        xlims = [np.amin(sim.bandit.mean_list) - 1.5,np.amax(sim.bandit.mean_list) + 1.5]
    elif sim.alg.var_dict['algtype'].__name__[-4:] == 'Beta':
        xlims = [0,1]

    for i in range(sim.bandit.n_arms):
        sim.bandit.pullArm(i)

    # distribution function
    def npdf(x, mu, sigma):
        return 0.39894228 / sigma * 0.60653066**(((x-mu)/(sigma))**2)

    # the animation update function
    def _update(num, confidence):
        nonlocal ax
        plt.cla()

        # step the bandit
        sim.runStep(1, horizon)
        samples = 500
        xdata = np.tile(np.linspace(xlims[0], xlims[1], samples), (n_arms,1))
        if sim.alg.var_dict['algtype'].__name__[-5:] == 'Gauss':
            mu = np.tile(
                sim.bandit.U.reshape(n_arms,1),
                (1,samples))
            sigma = np.tile(
                np.sqrt(1.0/sim.bandit.T).reshape(n_arms,1),
                (1,samples))
            ydata = npdf(xdata, mu, sigma)

        elif sim.alg.var_dict['algtype'].__name__[-4:] == 'Beta':
            beta_a = np.tile(
                (sim.bandit.arm_reward+1).reshape(n_arms,1),
                (1,samples))
            beta_b = np.tile(
                (sim.bandit.T - sim.bandit.arm_reward + 1).reshape(n_arms,1),
                (1,samples))
            ydata = beta.pdf(xdata, beta_a, beta_b)

        ydata_max = np.amax(ydata, axis=1)
        ymax = np.maximum(3, np.amax(ydata_max)*1.1)
        v_line_max = np.minimum(ydata_max/ymax, 1)

        for i in range(n_arms):
            cmap = cmap_colors.sequential1[i]
            plt.plot(xdata[i], ydata[i], "-", color=cmap(0.5))
            ax.fill_between(xdata[i], 0, ydata[i],
                color=cmap(0.5),
                alpha=0.2,
                linewidth=0.5,
                label="Arm {}: Pulls: {}".format(i+1, sim.bandit.T[i]))
            plt.axvline(sim.bandit.U[i],
                ymax=v_line_max[i],
                ls='dashed',
                color=cmap(0.8))
            plt.axvline(sim.bandit.mean_list[i],
                ymax=1,
                ls='dashdot',
                alpha=0.6,
                color=cmap(0.8))

        ylims = [0, ymax]
        # general formatting
        ax.grid(True)
        ax.set_ylim(ylims)
        ax.set_xlim(xlims)

        legend = plt.legend(loc='upper right')
        plt.gca().add_artist(legend)
        legend.get_frame().set_linewidth(1)

        # additional information legend
        blank_path_1 = mpatches.Patch(
            label='Regret: {:04.2f}'.format(sim.bandit.giveRegret()))
        blank_path_2 = mpatches.Patch(
            label='Timestep: {:2d}'.format(sim.bandit.timestep[0]))
        legend_2 = plt.legend(
            handles=[blank_path_1, blank_path_2],
            loc='upper left',
            handlelength=0,
            handletextpad=0)
        legend_2.get_frame().set_linewidth(1)

        plt.title(title)

    # the animation declaration
    my_ani = animation.FuncAnimation(fig, _update,
        fargs=(sim.bandit.U_conf, ),
        interval=2000/(core_dict['FPS']))  # fargs is used as the data required in update_hist

    plt.show()


import sys
import multiprocessing as mp
import re

import numpy as np
import yaml
from datetime import datetime
from pprint import pprint

from .formatting import bcolors

def _msplit(text, *limiters):
    """
    splits 'text' at every occurrence of limiter
    """
    for limit in limiters:
        text = text.replace(limit, '$!!7=4+[4}|[2')
    return text.split('$!!7=4+[4}|[2')


def Parse(user_file, **arg_dict):
    """
    Loads the user_file using YAML and then checks the dictionary for
    consistency. Once Parse is completed, it either exits with errors or
    returns a checked core_dict.
    """

    # Try to load the file and catch YAML syntax errors for nicer formatting
    try:
        # merge the two dicts
        core_dict = CoreDict(yaml.load(open(user_file, 'r')), **arg_dict)
    except yaml.parser.ParserError as e:
        # pull some information out of the ParserError
        location = (str(e).split("\n")[1].split(",")[1] + str(e).split("\n")[1].split(",")[2])[1:]
        e_msg = str(e).split("\n")[2]
        sys.exit(bcolors.FAIL + "\nYAML Syntax Error " + bcolors.ENDC
            + "in {}:\n  - {}\n".format(location, e_msg)
        )
    except yaml.scanner.ScannerError as e:
        # pull some information out of the ScannerError
        location = (str(e).split("\n")[1].split(",")[1]
            + str(e).split("\n")[1].split(",")[2]
            )[1:]
        e_msg = str(e).split("\n")[0]
        sys.exit(bcolors.FAIL + "\nYAML Syntax Error " + bcolors.ENDC
            + "in {}:\n  - {}\n".format(location, e_msg)
            )

    # creates a simulation dictionary list for easier access
    core_dict['sim'] = []
    for sim_key in list(core_dict.keys()):
        if sim_key[:10] == 'Simulation':
            core_dict['sim'].append(CoreDict(core_dict[sim_key]))
            del(core_dict[sim_key])

    # perform the dictionary check, and get the errors
    core_dict, errors = DictCheck(core_dict)

    # if there are errors, print them
    if errors:
        sys.exit(bcolors.FAIL
            + "\n+" + " ERROR LIST ".center(100, "-") + "+\n" + bcolors.ENDC
            + self.errors + "\n\n")
    else:
        print(bcolors.OKGREEN + "\n" + "No Errors!".center(100, "-") + "\n" + bcolors.ENDC)

    # post processing
    core_dict.post()

    # done!
    return core_dict


def DictCheck(core_dict):
    """
    The checking function, uses _check class to perform these checks nicely.

    Defaults are built first, followed by checks based on each init file and
    other specifics. The full list of checking functions with explanations
    can be found in the _check class.
    """
    check = _check(core_dict)
    if core_dict['InputData']:
        check.SimExist('label')
        check.Save()

    elif core_dict['init'] == 'Variable':
        check.Save()

        check.Conflict('horizon')
        check.Conflict('cycles')
        check.SimExist('label')

        check.Args()

    elif core_dict['init'] == 'Histogram':
        check.Save()

        check.Conflict('horizon')
        check.Conflict('cycles')
        check.SimExist('label')

        check.Bins()
    elif core_dict['init'] == 'Visualize':
        pass

    else:
        sys.exit(bcolors.FAIL
            + '\n\nIf you get this message, something wend badly wrong...\n'
            + bcolors.ENDC
        )
    return check.give()


class CoreDict(dict):
    """
    Subclasses dict to change behaviour with missing key (allows setting of deafults easily) and
    some other custom methods.

    TODO: read defaults from some external file, maybe with type-specific defaults
    """
    def __init__(self, in_dict, **arg_dict):
        dict.__init__(self, in_dict)
        for k, v in list(arg_dict.items()):  # removes unspecified arguments
            if v == None:
                del(arg_dict[k])
        try:
            default_txt = yaml.load(open("/".join(__file__.split('/')[:-1]) + "/defaults.txt", 'r'))
        except FileNotFoundError:
            sys.exit("oh")
        self.default = {**default_txt, **arg_dict}  # add the arg_dict arguments to overwrite any defaults
        self.ignore = {'InputData', 'DataFolder', 'Animate'}
        self.warning_list = []

    def __missing__(self, key):
        try:
            if key not in self.ignore:
                self.warning_list += ["'{} missing, defaulting to '{}'".format(
                    key,
                    self.default[key])]
            return self.default[key]
        except KeyError:
            raise KeyError("CoreDict: '{}' key missing in both the dict and the default dict."
                .format(key))

    def post(self):
        """
        Additional changes to be done after checking
        """
        self.default.setdefault('DataFolder', "{}".format(
            datetime.strftime(datetime.now(), '%Y-%m-%d %H_%M_%S')))
        try:
            if self.default['data']:
                self.default['DataFolder'] = self.default['data'] + "/" + self.default['DataFolder']
            if self.default['out']:
                self['PlotSave'] = self.default['out'] + "/" + self['PlotSave']
                self.default['PlotSave'] = self.default['out'] + "/" + self.default['PlotSave']
        except KeyError:
            pass
        if self['init'] == 'Histogram':
            self['total_lines'] = sum(sub_dict['cycles'] for sub_dict in self['sim'])
        elif self['init'] == 'Variable':
            self['total_lines'] = len(self['arg_list']) * len(self['sim'])

        if not self['PlotTitle']:
            # self['PlotTitle'] = self['PlotSave'].split("/")[-1].split(".")[0] + "\n"
            self['PlotTitle'] = _msplit(self['PlotSave'], "/", ".")[-1] + "\n"
        else:
            self['PlotTitle'] += "\n"


        return None


    def warnings(self):
        if not self.warning_list:
            return (bcolors.OKGREEN + "\n" + " No Warnings! ".center(100, "-") + "\n" + bcolors.ENDC)
        else:
            # removes repeats from the warning list
            return (bcolors.WARN
                + "\n+" + " WARNING LIST ".center(100, "-") + "+\n" + bcolors.ENDC
                + "".join(["- {}\n".format(item) for item in list(set(self.warning_list))]))
    def state(self):
        print(bcolors.OKBLUE + "\n" + " Completed Core Dict ".center(100, "-") + "\n" + bcolors.ENDC)
        pprint(self)
        print(bcolors.OKBLUE + "\n" + " Default Dict ".center(100, "-") + "\n" + bcolors.ENDC)
        pprint(self.default)
        print(bcolors.OKBLUE + "\n" + " End State ".center(100, "-") + "\n" + bcolors.ENDC)

        return None

class _check:
    """
    A class which checks a dictionary for errors. It contains a large number of
    methods which are convenient / reusable when parsing dictionary objects.

    """

    def __init__(self, core_dict):
        self.core_dict = core_dict
        self.errors = []

    def give(self):
        return (self.core_dict, "\n".join(set(self.errors)))  # removes repeats from self.errors

    def Visual(self, name):
        if core_dict['visual'] == 'confidence':
            pass

        elif core_dict['visual'] == 'ellipse':
            pass
        elif core_dict['visual'] == 'distribution':
            pass
        else:
            pass

    def Conflict(self, name):
        """
        Checks if 'name' exists in the sim dictionaries and in the overall
        dictionary; looks for conflict / multiple declarations of name. If
        there are no errors, it moves every instance of the declaration to
        the simulation subdictionaries for easier access later.
        """
        checklist = \
            {'low_{}'.format(name) : [False] * len(self.core_dict['sim']),
            'top_{}'.format(name) : [False]}

        if name in self.core_dict.keys():
            checklist['top_{}'.format(name)] = True

        for i, sub_dict in enumerate(self.core_dict['sim']):
            if name in sub_dict.keys():
                checklist['low_{}'.format(name)][i] = True

        if not checklist['top_{}'.format(name)]:
            cntr = 0
            for item in checklist['low_{}'.format(name)]:
                if not item:
                    cntr += 1
            if cntr != 0:
                self.errors += ["- ({0}) was not declared at the top level, and you are missing a"
                    " ({0}) declaration in ({1}) simulation(s)".format(name, cntr)]
        else:
            cntr = 0
            for item in checklist['low_{}'.format(name)]:
                if item:
                    cntr += 1
            if cntr != 0:
                self.errors += ["- ({0}) was declared at the top level, but "
                "you have ({0}) declarations in ({1}) simulation(s).".format(
                    name, cntr)]
            else:
                for sub_dict in self.core_dict['sim']:
                    sub_dict['{}'.format(name)] = \
                        self.core_dict['{}'.format(name)]
                del self.core_dict['{}'.format(name)]
        return None

    def Exist(self, name):
        """
        Checks if 'name' exists in the core_dict
        """
        if name not in self.core_dict.keys():
            self.errors += ["- You are missing a top-level ({}) declaration.".format(
                name)]
        else:
            pass
        return None


    def SimExist(self, name):
        """
        Checks if 'name' exists in every single simulation sub_dictionary, and
        adds an error if it doesn't
        """
        checklist = {'low_{}'.format(name) : [False] * len(self.core_dict['sim'])}
        for i, sub_dict in enumerate(self.core_dict['sim']):
            if name in sub_dict.keys():
                checklist['low_{}'.format(name)][i] = True
        if (checklist['low_{}'.format(name)]
            != [True] * len(self.core_dict['sim'])):
            self.errors += ["- You are missing at least one simulation ({}) declaration.".format(
                name)]
        return None


    def Save(self):
        """
        Checks the save format for .pdf or .png. It returns an eror otherwise.
        """
        if (not self.core_dict['PlotSave'].endswith(".pdf")
            and not self.core_dict['PlotSave'].endswith(".png")):
            self.errors += ["- PlotSave: You can only save a plot as a '.pdf' or '.png'. PDF is "
                "the recommended file type for image quality."]
        return None


    def Bins(self):
        """
        Determines how many bins the histogram should have based on the proper-
        ties of the bandit.
        """
        self.core_dict['bins'] = []
        for sub_dict in self.core_dict['sim']:
            if sub_dict['Bandit']['ArmList'][0][0] == 'Linear':
                mean_list = [np.inner(arm[1], sub_dict['Bandit']['MeanVector'])
                    for arm in sub_dict['Bandit']['ArmList']]
            else:
                mean_list = [arm[1] for arm in sub_dict['Bandit']['ArmList']]
            self.core_dict['bins'].append(np.amax(
                mean_list - np.amin(mean_list)) * sub_dict['horizon'])
        return None


    def Args(self):
        """
        When using the Variable class, args can be declared in multiple ways.
        This builds a single list of args out of the inputs to be used later.
        """
        checklist = {'linspace': False, 'args': False}
        if ('domain' in self.core_dict['Var'].keys()
            and 'samples' in self.core_dict['Var'].keys()):
            checklist['linspace'] = True

        if 'arg_list' in self.core_dict['Var'].keys():
            checklist['args'] = True

        if checklist['linspace'] and checklist['args']:
            self.errors += ["- Declare either (domain and samples) or (args), not both."]

        elif not checklist['linspace'] and not checklist['args']:
            self.errors += ["- You must declare either (domain and samples) or (args)."]
        else:
            try:
                self.core_dict['arg_list'] = np.linspace(
                    self.core_dict['Var']['domain'][0],
                    self.core_dict['Var']['domain'][1],
                    self.core_dict['Var']['samples'])
            except KeyError:
                pass
        return None



import numpy as np
from numpy.linalg import inv


def random(bandit, var_dict):
    """
    Choose arms randomly.

    Arm type support:
        * Bernoulli
        * Normal

    Visualization support:
        *
    """
    return np.random.randint(bandit.n_arms)


def greedy(bandit, var_dict):
    """
    Choose arms greedily.

    Arm type support:
        * Bernoulli
        * Normal

    Visualization support:
        *
    """
    return np.argmax(bandit.U)


def greedy_ep(bandit, var_dict):
    """
    Random with probability epsilon, else greedy.

    Arm type support:
        * Bernoulli
        * Normal

    Visualization support:
        *
    """
    if np.random.random() < var_dict['epsilon']:
        random(bandit, var_dict)
    else:
        greedy(bandit, var_dict)



def UCB(bandit, var_dict):
    """
    Upper Confidence Bound algorithm.

    Arm type support:
        * Bernoulli
        * Normal

    Visualization support:
        * ConfAnimation
    """
    index_type = var_dict['incr']
    alpha = var_dict['alpha']

    # now everything is a vector, and numpy does the operations to all of them at the same time
    bandit.U_conf = bandit.U + np.sqrt(alpha / bandit.T * np.log(index_type(bandit)))
    return(np.argmax(bandit.U_conf))



def UCB_KL(bandit, var_dict):
    """
    KL Upper Confidence Bound algorithm.

    Arm type support:
        * Bernoulli

    Visualization support:
        * ConfAnimation
    """

    # TODO find a way to vectorize this...
    # def v_f(x):  # vector version of f(x)
    #   zero_index = bandit.U == 0
    #   x[zero_index] = np.log(1/(1-x))
    #   x[invert(zero_index)] = bandit.U * np.log(bandit.U/x) + (1-bandit.U) * np.log((1-bandit.U)/(1-x)) - target
    #   return x
    def kl( p,q ):  # the KL entropy form; specified for p = 0 since numpy cannot properly evaluate the limit
        if p == 0:
            return( np.log(1/(1-q)))
        else:
            return( p * np.log(p/q) + (1-p) * np.log((1-p)/(1-q)) )

    def dkl( p,q ):  # the derivative of the KL entropy form with respect to q
        return( (q - p) / (q * (1-q)) )

    index_type = var_dict['incr']
    precision = 1.0/bandit.horizon[0]

    for i, (num, pulls, index) in enumerate(zip(bandit.U, bandit.T, index_type(bandit))):
        target = 1.0/pulls * np.log(index)
        if num == 1:
            bandit.U_conf[i] = 1
        else:
            if bandit.U_conf[i] == 1:
                bandit.U_conf[i] = num
            while True:
                if kl(num, bandit.U_conf[i]) < target:
                    bandit.U_conf[i] = (bandit.U_conf[i] + 1) * 0.5
                else:
                    break
            n = 1
            while n > precision:
                n = (kl(num, bandit.U_conf[i]) - target)/dkl(num, bandit.U_conf[i])
                bandit.U_conf[i] = bandit.U_conf[i] - n
    return np.argmax(bandit.U_conf)


def Bayes_Gauss(bandit, var_dict):
    """
    Bayes Gaussian Algorithm

    Note: for precision reasons, only accurate to a horizon of 5000!

    Arm type support:
        * Bernoulli
        * Normal

    Visualization support:
        * DistAnimation  # TODO - no support yet!
    """
    def npdf(x, mu, sigma):
        return 0.39894228 / sigma * 0.60653066**(((x-mu)/(sigma))**2)
    def ncdf(x, mu, sigma):
        return(1.0 / (1+np.e**(-0.07056*((x-mu)/(sigma))**3 - 1.5976*((x-mu)/(sigma)))))
    index_type = var_dict['incr']

    precision = 1.0 / bandit.horizon[0]
    for i, (num, pulls, index) in enumerate(zip(bandit.U, bandit.T, index_type(bandit))):

        if pulls == 0:
            bandit.U_conf[i] = 100000

        init = num
        mu = num
        sigma = np.sqrt(1.0/pulls)

        n = 1
        while n > precision or n * (-1) > precision:
            n = (ncdf(init, mu, sigma) - (1-1/index))/npdf(init, mu, sigma)
            init = init - n

        bandit.U_conf[i] = init

    return np.argmax(bandit.U_conf)

def TS_Beta(bandit, var_dict):
    """
    Beta Thompson Sampling algorithm

    Arm type support:
        * Bernoulli

    Visualization support:
        * DistAnimation  # TODO - no support yet!
    """
    return np.argmax(np.random.beta(bandit.arm_reward + 1, bandit.T - bandit.arm_reward + 1))

def TS_Gauss(bandit, var_dict):
    """
    Gaussian Thompson Sampling algorithm

    Arm type support:
        * Normal

    Visualization support:
        * DistAnimation  # TODO - no support yet!
    """
    mask = bandit.T == 0
    bandit.U_conf[mask] += 100000
    bandit.U_conf[np.invert(mask)] = np.random.normal(
        bandit.U,
        np.sqrt(1.0/bandit.T))
    return np.argmax(bandit.U_conf)


def UCB_Lin(bandit, var_dict):
    """
    Finite Linear Upper Confidence Bound algorithm

    Arm type support:
        * Linear

    Visualization support:
        * EllipseAnimation
    """
    rho = bandit.dim * np.log(bandit.timestep[0])
    bandit.U_conf = (np.inner(bandit.arm_vecs, bandit.U)
        + np.sqrt(rho * np.einsum('ij,ij->i',
            np.inner(bandit.arm_vecs,
            inv(bandit.G)), bandit.arm_vecs)))

    return np.argmax(bandit.U_conf)

def TS_Lin(bandit, var_dict):
    """
    Does Thompson Sampling from the multivariate distribution with mean
    bandit.U and covariance inv(bandit.G)
    """
    bandit.U_conf = np.einsum('ij,ij->i',
        np.random.multivariate_normal(
            bandit.U,
            inv(bandit.G),
            bandit.n_arms),
        bandit.arm_vecs)

    return np.argmax(bandit.U_conf)

def B1(bandit):
    return bandit.timestep

def B2(bandit):
    return bandit.horizon

def B3(bandit):
    return bandit.horizon / bandit.timestep

def B4(bandit):
    return bandit.timestep / bandit.T

def B5(bandit):
    return bandit.horizon / bandit.T

def B6(bandit):
    """
    * Makes a n * m matrix with every element sqrt(T_n / T_m)
    * Takes the minimum of that matrix and a matric of ones, element by element
    * Takes the sum down columns
    * Multiplies by original T element by element across rows
    * Divides the timestep by each element
    * adds 1
    """
    return 1 + bandit.timestep / (bandit.T * np.sum(
        np.minimum(
            np.ones((bandit.n_arms, bandit.n_arms)),
            (np.tile(np.array([bandit.T]).T, (1, bandit.n_arms))/bandit.T)),
        0))

def B7(bandit):
    """
    * Makes a n * m matrix with every element sqrt(T_n / T_m)
    * Takes the minimum of that matrix and a matric of ones, element by element
    * Takes the sum down columns
    * Multiplies by original T element by element across rows
    * Divides the timestep by each element
    * Takes the maximum of each element with respect to e
    """
    return np.maximum(np.e, bandit.timestep / (bandit.T * np.sum(
        np.minimum(
            np.ones((bandit.n_arms, bandit.n_arms)),
            (np.tile(np.array([bandit.T]).T, (1, bandit.n_arms))/bandit.T)),
        0)))

def B8(bandit):
    return bandit.horizon / (bandit.n_arms * bandit.T)

def B9(bandit):
    return bandit.timestep / (bandit.n_arms * bandit.T) + 1



from setuptools import setup

setup(name='banditvis',
      version='0.1',
      description='A library for simulating bandit algorithms.',
      url='https://github.com/alexrutar/banditvis',
      author='Alex Rutar',
      author_email='arutar@uwaterloo.ca',
      license='MIT',
      packages=['banditvis'],
      install_requres=['numpy', 'pyyaml', 'scipy', 'matplotlib'],
      zip_safe=False)


import multiprocessing as mp
import threading as th
from pprint import pprint
import sys
import time
import signal

from .parse import Parse
from .plot import *
from .data import *
from .animation import *
from .formatting import bcolors

def _checkInput(**arg_dict):
    """
    Gets the input from the user commmand and checks it for potential errors.
    """

    if arg_dict['input'].endswith('.txt'):
        pass
    else:
        sys.exit("ERROR: You need to input a .txt file.")
    try:
        input_file = arg_dict['input']
        del(arg_dict['input'])
        core_dict = Parse("{}".format(input_file), **arg_dict)
    except FileNotFoundError:
        sys.exit("ERROR: The file '{}' you tried to input doesn't exist in the current directory.".format(input_file))
    return core_dict

def _init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class _statusThread(th.Thread):
    def __init__(self, core_dict):
        super(_statusThread, self).__init__()
        self.core_dict = core_dict
        self.n_files = len(core_dict['sim'])
        self.daemon = True
        self.total = core_dict['total_lines']
        self.processes = core_dict['Multiprocess']
    def run(self):
        while True:  # TODO fix this to have a termination
            time.sleep(0.5)
            current = [0] * self.n_files
            try:
                for i in range(self.n_files):
                    with open("{}/data{}.txt".format(self.core_dict['DataFolder'], i), 'r') as file:
                        current[i] += sum(1 for line in file) - 1
            except FileNotFoundError:
                pass
            sys.stdout.write("\r" + " {:2.0f} % complete ".format(sum(current)*100/self.total).center(100, "-"))
            sys.stdout.flush()


def run(**arg_dict):
    core_dict = _checkInput(**arg_dict)
    pool = mp.Pool(core_dict['Multiprocess'], _init_worker)
    core_dict.state()

    try:
        if core_dict['init'] == 'Histogram':
            status = _statusThread(core_dict)
            status.start()
            if not core_dict['InputData']:
                HistData(core_dict, pool)
                pool.close()
                pool.join()
            if core_dict['Animate']:  # TODO in general fix this thing
                while True:
                    try:
                        HistAnimation(core_dict)
                        break
                    except FileNotFoundError:
                        time.sleep(0.3)
            HistPlot(core_dict)
        elif core_dict['init'] == 'Variable':
            status = _statusThread(core_dict)
            status.start()
            if not core_dict['InputData']:
                proc_list = VarData(core_dict)
            _run(proc_list, n_processes)
            VarPlot(core_dict)

        elif core_dict['init'] == 'Visualize':
            if core_dict['visual'] == 'ellipse':
                EllipseAnimation(core_dict)
            elif core_dict['visual'] == 'confidence':
                ConfAnimation(core_dict)
            elif core_dict['visual'] == 'distribution':
                DistAnimation(core_dict)

        print("\n\n" + bcolors.OKGREEN + " Done! ".center(100,"-") + bcolors.ENDC + "\n")
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
        print("\r" + " "*150)
        sys.exit("\n\n" + bcolors.FAIL + " Keyboard Interrupt! ".center(100,"-") + bcolors.ENDC + "\n")





import numpy as np
from numpy.linalg import inv


class Bandit:

    """
    The basic Bandit class.

    The Bandit class defines the bandit itself. As an argument, it takes a list
    of arm objects, defining the behaviour of the environment. When the
    Bandit.pullArm(n) method is called, the n^th arm (indexed from 0) is
    pulled, updating various information stored within the bandit. This
    information is stored as follows:

        * Bandit.arms : list of arms objects within the bandit, which contains:
            * Bandit.arms[n].mean : the mean of the n^th arm
            * Bandit.arms[n].info : info about the n_th arm
        * Bandit.n_arms : number of arms that the Bandit has
        * Bandit.T : a vector storing the number of pulls of each arm
        * Bandit.U : a vector storing the cumulative average of each arm
        * Bandit.U_conf : the adjusted version of U [n] taking into account
            some confidence interval (not always used)
        * Bandit.timestep[0] : the number of timesteps that have passed so far
            (incremented with every arm pull)
            * Note : Bandit.timestep is a vector for use in algorithms
        * Bandit.total_reward : the total reward that the Bandit has
            accumulated so far
        * Bandit.arm_reward : the reward that each arm has so far
        * Bandit.reward : the reward that the Bandit recieved after the
            previous timestep

    The Bandit class also contains an interactive method called by
    Bandit.pullArm(n). The algorithm does the following:

        * pulls the n^th arm
        * increments T[n]
        * increments timestep
        * updates U[n] in a linear fashion (so that U[n] is simply an
            unweighted average of all previous rewards)
        * increments total_reward and arm_reward[n] to store the information

    The method Bandit.reset() reverts the Bandit to the original state in which
    it started, while still storing the arms; it resets tracking information.
    There is also a method called by Bandit.fullInfo(), which returns a large
    amount of information about the bandit. This is mostly good for trouble-
    shooting and examining the behaviour of the bandit in detail. Specifically,
    it gives the following information:

        * for each arm :
            * the mean of the distribution
            * the current system average
            * the number of pulls
            * confidence
        * the number of arms
        * the timestep
        * the horizon
        * the total reward
    """

    def __init__(self, arm_type_object):
        self.arms = arm_type_object
        self.n_arms = len(arm_type_object)
        self.T = np.zeros(self.n_arms, dtype=np.int)
        self.U = np.zeros(self.n_arms)
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        self.total_reward = 0
        self.arm_reward = np.zeros(self.n_arms, dtype=np.int)

        self.mean_list = np.array([self.arms[arm].mean
            for arm in range(self.n_arms)])



    def giveRegret(self):
        self.regret = 0
        for arm in range(self.n_arms):
            self.regret += self.T[arm] * (np.amax(self.mean_list) - self.arms[arm].mean)

        return(self.regret)


    def pullArm(self, arm):
        self.T[arm] += 1
        self.timestep += 1

        self.reward = self.arms[arm].pull()

        self.U[arm] = self.U[arm] + 1/self.T[arm] * (self.reward - self.U[arm])

        self.total_reward += self.reward
        self.arm_reward[arm] += self.reward

        return None


    def reset(self):  # resets everything in the bandit for re-use!
        self.T = np.zeros(self.n_arms, dtype=np.int)
        self.U = np.zeros(self.n_arms)
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        self.total_reward = 0
        self.arm_reward = np.zeros(self.n_arms, dtype=np.int)
        self.regret = 0

        return None


    def fullInfo(self):
        print("\n" + "+" + "-"*85 + "+")

        for arm in range(self.n_arms):
            print("| Arm {0}: mean ({1}) average reward ({2:f}), confidence "
                "({3:f}), called ({4}) times"
                .format(
                    arm,
                    self.arms[arm].mean,
                    self.U[arm],
                    self.U_conf[arm],
                    self.T[arm]))
        print("|\n| {0} arms, total reward ({1}), timestep ({2} / {3})"
            .format(
                self.n_arms,
                self.total_reward,
                self.timestep[0],
                self.horizon[0]))

        print("+" + "-"*85 + "+" + "\n")
        return None




class LinBandit:
    """
    The Linear Bandit class is very similar to the original Bandit class, in
    that it contains similar methods which do the same thing as the original
    Bandit. However, it differs in the way in which is stores averages, and
    takes different arm arguments.

    It takes a list of Linear arms and a vector mean to initialize.

    If the normalized option is True, it preserves vector direction but makes
    each vector unit length. It stores the following information internally:
        * self.normalized: boolean for normalized
        * self.mean: the vector mean
        * self.n_arms: the number of arms
        * self.T: the number of pulls of each arm
        * self.arms: the arm_object_list input
        * self.dim: dimension
        * self.arm_vecs: a list of the arm vectors
        * self.G: the gram matrix, initialized with the identity to ensure
            invertibility
        * self.U: the system average vector
        * self.A: the sum of arm vectors chosen multiplied by the scalar reward
            it recieved
        * self.U_conf: the confidence value for each arm
        * self.timestep: the timestep
    """
    def __init__(self, arm_object_list, vector_mean, normalized=False):
        self.normalized = normalized
        self.mean = vector_mean
        self.n_arms =  len(arm_object_list)
        self.T = np.zeros(len(arm_object_list), dtype = int)
        self.arms = arm_object_list
        self.dim = arm_object_list[0].dim
        self.arm_vecs = np.array([arm.arm_vec for arm in self.arms])
        self.G = np.identity(self.dim)  # .dim is the dimension
        self.U = np.zeros(self.dim)  # vector of averages
        self.A = np.zeros(self.dim)  # vectors of actions taken so far weighted by the reward
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        for arm in arm_object_list:
            arm.mean_vec = vector_mean
        if self.normalized:
            self.mean = self.mean/np.linalg.norm(self.mean)
            self.arm_vecs = np.array([arm/np.linalg.norm(arm) for arm in self.arm_vecs])
    def pullArm(self, arm):
        self.T[arm] += 1
        self.timestep += 1  # update timesetp

        # self.G is the sum of the products X X^T of the arm pulled in each round X
        self.G += np.outer(self.arm_vecs[arm], self.arm_vecs[arm])  # self.arm_vecs[arm] is the arm vector for a given [arm]

        # self.A is the sum of the arm vector pulled so far, scaled by the reward each pull resulted in
        self.A += self.arm_vecs[arm] * self.arms[arm].pull()

        # self.U is the "most likely" value of the mean vector; it is calculated based on least squares
        # update the reward approximation based on the inverse of G times the scaled sum of chosen arms
        self.U = np.dot(inv(self.G), self.A)

        return None

    def giveRegret(self):
        return (np.amax(np.dot(self.arm_vecs, self.mean)) * self.timestep[0] - np.dot(
            self.T,
            np.dot(self.arm_vecs, self.mean))
        )

    def reset(self):
        self.T = np.zeros(self.n_arms)
        self.G = np.identity(self.dim)  # .dim is the dimension
        self.U = np.zeros(self.dim)  # vector
        self.A = np.zeros(self.dim)  # vectors of actions taken so far weighted by the reward
        self.U_conf = np.zeros(self.n_arms)
        self.timestep = np.zeros(self.n_arms, dtype=np.int)
        return None

    def fullInfo(self):
        print("\n" + "+" + "-"*105 + "+")
        print("| Mean Vector: {}\n|".format(self.mean))
        print("| Gram Matrix: \n{}\n| ".format(self.G))
        for arm in range(self.n_arms):
            print("| Arm {0}: {1}, confidence ({2:f}), called ({3}) times\n| ".format(
                arm,
                self.arms[arm].arm_vec,
                self.U_conf[arm],
                self.T[arm])
            )
        print("Mean Reward Vector: {}".format(self.U))
        print("|\n| {0} arms,  timestep ({1} / {2})".format(
            self.n_arms,
            self.timestep[0],
            self.horizon[0])
        )
        print("+" + "-"*105 + "+" + "\n")
        return None

class Algorithm:
    """
    The Algorithm class specifies the behaviour of an algorithm.

    As an argument, it is passed a dictionary containing an 'algtype' key along
    with other keys.

    The Algorithm.giveArm() chooses an arm based on the internal state of the
    bandit or other factors. For a list of the supported algorithms and more
    detailed information, see the algorithm_types module.
    """
    def __init__(self, **variables):
        self.var_dict = variables

    def giveArm(self):
        return(self.var_dict['algtype'](self.bandit, self.var_dict))
